import json
import os

import typer


def write_config(config: dict):
    config_dir = os.path.expanduser("~/.cybersharing")
    if not os.path.exists(config_dir):
        os.makedirs(config_dir)

    config_path = os.path.join(config_dir, "config.json")
    with open(config_path, "w") as f:
        f.write(json.dumps(config))


def read_config() -> dict:
    try:
        config = os.path.join(os.path.expanduser("~/.cybersharing"), "config.json")
        with open(config, "r") as f:
            return json.loads(f.read())
    except FileNotFoundError:
        return {}
    except json.JSONDecodeError:
        typer.echo("Invalid config file")
        raise typer.Abort()



import binascii
import datetime
import requests
import typer
import re
import os
from typing import List
from base64 import b64encode
import json

from cybersharing.api import create_container, download_file, get_container_details, get_username, preupload_file
from cybersharing.config import read_config, write_config
from cybersharing.encryption import decrypt_string, generate_key

main = typer.Typer()


def validate_auth():
    config = read_config()
    username = config.get("username")
    if username:
        try:
            new_username = get_username(read_config()["token"])
        except requests.exceptions.HTTPError:
            typer.echo("Authentication failed. Please logout or login again.")
            raise typer.Abort()

        if new_username != username:
            write_config({
                "token": config["token"],
                "username": new_username
            })

        typer.echo(f"Logged in as {new_username}")


@main.command()
def logout():
    """
    Logout from the API.
    """

    write_config({})
    typer.echo("Logged out successfully")


@main.command()
def login(token: str):
    """
    Login to the API.
    """

    try:
        username = get_username(token)
    except requests.exceptions.HTTPError as e:
        typer.echo("Invalid token")
        raise typer.Abort()

    config = {
        "token": token,
        "username": username
    }

    write_config(config)
    typer.echo("Logged in successfully")


@main.command()
def upload(files: List[str],
           password: str = typer.Option(None, help="Password for your files"),
           encrypt: bool = typer.Option(False, help="Whether to encrypt the files"),
           padding: int = typer.Option(None, help="Padding in MiB for encryption"),
           description: str = typer.Option(None, help="Description of the upload"),
           extra_long_url: bool = typer.Option(False, help="Generate an extra long URL"),
           public: bool = typer.Option(False, help="Add the upload to your public profile"),
           save_to_history: bool = typer.Option(False, help="Save the upload to the IP history"),
           permanent: bool = typer.Option(False, help="Disable expiration date"),
           expiration_date: datetime.datetime = typer.Option(datetime.datetime.now(
           ) + datetime.timedelta(days=7), help="Expiration date, defaults to 7 days from now"),
           expiration_days: int = typer.Option(None, help="Relative expiration time in days"),
           expiration_hours: int = typer.Option(None, help="Relative expiration time in hours"),
           expiration_minutes: int = typer.Option(None, help="Relative expiration time in minutes"),
           generate_password: bool = typer.Option(False, help="Generate a random secure password"),
           auth_check: bool = typer.Option(True, help="Check if the user is authenticated before uploading")
           ):
    """
    Upload files.
    """

    if auth_check:
        validate_auth()

    if generate_password:
        password = b64encode(os.urandom(32)).decode().rstrip("=")
        typer.echo(f"Generated password: {password}")

    if encrypt and not password:
        password = typer.prompt("Password", hide_input=True)
        if not password:
            typer.echo("Password is required for encryption")
            raise typer.Abort()

    encryption_salt, encryption_key = generate_key(password) if encrypt else (None, None)

    for file in files:
        if not os.path.exists(file):
            typer.echo(f"File {file} not found")
            raise typer.Abort()

    max_filename_length = max([len(os.path.basename(file)) for file in files])

    if expiration_days or expiration_hours or expiration_minutes:
        expiration_date = datetime.datetime.now()
        if expiration_days:
            expiration_date += datetime.timedelta(days=expiration_days)
        if expiration_hours:
            expiration_date += datetime.timedelta(hours=expiration_hours)
        if expiration_minutes:
            expiration_date += datetime.timedelta(minutes=expiration_minutes)

    authenticated = read_config().get("username") is not None
    if expiration_date < datetime.datetime.now() + datetime.timedelta(minutes=5):
        typer.echo("Minimum expiration time is 5 minutes")
        raise typer.Abort()
    elif not authenticated and (expiration_date > datetime.datetime.now() + datetime.timedelta(days=7) or permanent):
        typer.echo("Maximum expiration time for anonymous users is 7 days. Please login to remove this limit.")
        raise typer.Abort()

    if permanent:
        expiration_date = None

    preupload_ids = []
    for file in files:
        preupload_id = preupload_file(file, max_filename_length, encryption_key, encrypt, padding)
        preupload_ids.append(preupload_id)

    fragment = create_container(preupload_ids, password, encryption_salt, encrypt, description, extra_long_url, public, save_to_history, expiration_date)
    typer.echo(f"Access your files at https://cybersharing.net/s/{fragment}")


@main.command()
def download(url: str,
             dest: str = typer.Option(".", help="Destination directory for the downloaded files"),
             password: str = typer.Option(None, help="Password for your files"),
             auth_check: bool = typer.Option(True, help="Check if the user is authenticated before downloading")
             ):
    """
    Download a file from a URL.
    """

    if auth_check:
        validate_auth()

    if not re.match(r"https://cybersharing.net/s/[0-9a-f]+", url):
        typer.echo("Invalid URL")
        raise typer.Abort()

    fragment = url.split("/")[-1]
    details, password = get_container_details(fragment, password)

    uploads = details["uploads"]
    total_files = len(uploads)
    dest = os.path.abspath(dest)
    downloaded_files = 0

    encryption_key = generate_key(password, bytes.fromhex(details["encryptionSalt"]))[1] if details["isEncrypted"] else None

    if encryption_key:
        for upload in uploads:
            upload["fileName"] = decrypt_string(encryption_key, upload["fileName"])
            upload["contentType"] = decrypt_string(encryption_key, upload["contentType"])
            upload["fileSize"] = int(decrypt_string(encryption_key, upload["encryptedFileSize"]))

    max_filename_length = max([len(upload["fileName"]) for upload in uploads])
    total_size = sum([upload["fileSize"] for upload in uploads])

    for upload in uploads:
        download_file(dest, details["id"], upload["id"], details["signature"], upload["fileName"],
                      upload["fileSize"], max_filename_length, encryption_key, total_files, downloaded_files)
        downloaded_files += 1

    typer.echo(f"Downloaded {total_files} files ({total_size} bytes) to {dest}")


if __name__ == "__main__":
    main()



import io
import os
import hashlib
from typing import Tuple
from Crypto.Cipher import AES
from Crypto.Util.Padding import pad, unpad


def generate_key(password: str, salt: bytes = None) -> Tuple[bytes, bytes]:
    salt = os.urandom(16) if not salt else salt
    derived_key = hashlib.pbkdf2_hmac('sha256', password.encode(), salt, 10000, dklen=32)
    return (salt, derived_key)


def decrypt(key: bytes, iv: bytes, data: bytes) -> bytes:
    cipher = AES.new(key, AES.MODE_CBC, iv=iv)
    return cipher.decrypt(data)


def encrypt(key: bytes, iv: bytes, data: bytes) -> bytes:
    cipher = AES.new(key, AES.MODE_CBC, iv=iv)
    return cipher.encrypt(data)


def encrypt_string(key: bytes, data: str) -> str:
    iv = os.urandom(16)
    data = data.ljust(1024, "\0")
    return iv.hex() + encrypt(key, iv, pad(data.encode(), AES.block_size)).hex()


def decrypt_string(key: bytes, data: str) -> str:
    iv = bytes.fromhex(data[:32])
    data = bytes.fromhex(data[32:])
    return unpad(decrypt(key, iv, data), AES.block_size).decode().rstrip("\0")


def get_expected_encrypted_length(plaintext_length: int) -> int:
    return 16 + plaintext_length // 16 * 16 + 16

def get_expected_length_with_padding(plaintext_length: int, padding: int) -> int:
    return plaintext_length + padding - plaintext_length % padding if padding else plaintext_length


class EncryptedReader(io.IOBase):
    def __init__(self, file: io.BufferedReader, size: int, key: bytes, padding: int):
        self.key = key
        self.position = 0
        self.iv = None

        # source file
        self.file = file
        self.file_length = size

        self.file_length_with_padding = get_expected_length_with_padding(size, padding)
        self.padding_left = self.file_length_with_padding - size

        self.encrypted_length = get_expected_encrypted_length(self.file_length_with_padding)

        self.remaining_data = b""

    def read(self, size: int = -1):
        # if the last block wasn't fully read, add it to the beginning of the next read
        chunk = self.remaining_data
        self.remaining_data = b""
        size -= len(chunk)

        # size here is always larger than 16
        if not self.iv:
            self.iv = os.urandom(16)
            size -= 16
            chunk += self.iv

        # how many bytes of the last block should be sent
        remainder = size % 16

        block_position = self.position + size
        should_pad = block_position >= self.file_length_with_padding

        # the data will be processed based on this size
        ceiling_size = size - remainder + 16
        plaintext = self.file.read(ceiling_size)

        # pad with dummy data if necessary
        padding_amount = min(ceiling_size - len(plaintext), self.padding_left)
        self.padding_left -= padding_amount
        plaintext += b"\0" * padding_amount

        self.position += size

        if plaintext:
            chunk += encrypt(self.key, self.iv, pad(plaintext, AES.block_size) if should_pad else plaintext)

        self.iv = chunk[-32:-16] if remainder else chunk[-16:]
        self.remaining_data = chunk[-16+remainder:] if remainder else b""
        chunk = chunk[:-16+remainder] if remainder else chunk

        return chunk

    def seek(self, offset: int, whence: int = 0):
        assert offset == self.position, "seek only supports current position"
        self.position = offset
        return self.position

    def tell(self):
        return self.position

    def __len__(self):
        return self.encrypted_length



import binascii
import datetime
import mimetypes
import os
from typing import List
import requests
from requests_toolbelt import MultipartEncoder, MultipartEncoderMonitor
import typer
from cybersharing.config import read_config
from cybersharing.encryption import EncryptedReader, decrypt, encrypt_string, generate_key, get_expected_encrypted_length, get_expected_length_with_padding
from Crypto.Util.Padding import unpad
from Crypto.Cipher import AES

BASE_URL = "https://cybersharing.net"


def get_username(token: str) -> bool:
    response = requests.get(f"{BASE_URL}/api/users/me", headers={"X-CliToken": token})
    response.raise_for_status()
    return response.json()["userName"]


def get_headers() -> dict:
    config = read_config()
    return {"X-CliToken": config["token"]} if config.get("token") else {}


def create_container(preupload_ids: List[str], password: str, encryption_salt: bytes | None, is_encrypted: bool, description: str | None, extra_long_url: bool, is_public: bool, save_to_history: bool, expiration_date: datetime.datetime) -> str:
    url = f"{BASE_URL}/api/upload/create-container"
    unlock_password = "".join([x.hex() for x in generate_key(password)]) if password else None
    payload = {
        "preuploads": preupload_ids,
        "description": description,
        "encryptionSalt": encryption_salt.hex() if encryption_salt else None,
        "expirationDate": expiration_date.isoformat() if expiration_date else None,
        "extraLongUrl": extra_long_url,
        "isEncrypted": is_encrypted,
        "isPublic": is_public,
        "password": unlock_password,
        "saveToHistory": save_to_history,
    }
    response = requests.post(url, json=payload, headers=get_headers())
    response.raise_for_status()
    response_json = response.json()
    return response_json["pathFragment"]


def get_container_details(fragment: str, password: str = None, salt: bytes = None, iterations: int = 10000, first_attempt: bool = True):
    url = f"{BASE_URL}/api/containers/{fragment}"
    unlock_password = generate_key(password, salt)[1].hex() if password and salt else None
    payload = {"password": unlock_password}

    response = requests.post(url, json=payload, headers=get_headers())

    if response.status_code == 400:
        error = response.json()
        if error["error"] == "Invalid password":
            salt = bytes.fromhex(error["salt"])
            iterations = error["iterations"]
            if first_attempt and password:
                return get_container_details(fragment, password, salt, iterations, first_attempt=False)
            elif not password:
                password = typer.prompt("Password", hide_input=True)
                return get_container_details(fragment, password, salt, iterations, first_attempt=False)
            else:
                typer.echo("Invalid password")
                raise typer.Abort()

    response.raise_for_status()
    return response.json(), password


def preupload_file(file: str, pad_to: int, encryption_key: bytes = None, encrypt: bool = False, padding: int = None) -> str:
    padding = (0 if not padding else padding) * 1024 * 1024
    file_size = os.path.getsize(file)
    file_name = os.path.basename(file)
    content_type = mimetypes.guess_type(file_name)[0] or "application/octet-stream"

    encrypted_file_name = encrypt_string(encryption_key, file_name) if encrypt else None
    encrypted_content_type = encrypt_string(encryption_key, content_type) if encrypt else None
    encrypted_file_size = encrypt_string(encryption_key, str(file_size)) if encrypt else None

    preupload_size = file_size if not encrypt else get_expected_encrypted_length(
        get_expected_length_with_padding(file_size, padding))

    def upload_callback(monitor: MultipartEncoderMonitor):
        progress = min(monitor.bytes_read / preupload_size, 1)
        typer.echo(f"Uploading {file_name.ljust(pad_to, ' ')} [{int(progress * 20) * '█':20s}] {progress:.0%}\r", nl=False)

    with open(file, "rb") as f:
        fields = {
            "size": str(preupload_size),
            "file": (file_name, f, content_type)
        } if not encrypt else {
            "size": str(preupload_size),
            "encryptedSize": encrypted_file_size,
            "file": (encrypted_file_name, EncryptedReader(f, file_size, encryption_key, padding), encrypted_content_type)
        }
        encoder = MultipartEncoder(fields)
        monitor = MultipartEncoderMonitor(encoder, upload_callback)
        headers = {"Content-Type": monitor.content_type}
        headers.update(get_headers())
        response = requests.post("https://cybersharing.net/api/upload/pre-upload", data=monitor, headers=headers)
        typer.echo("")
        response.raise_for_status()
        response_json = response.json()
        return response_json["id"]


def download_file(dest: str, container_id: str, upload_id: str, signature: str, filename: str, file_size: int, pad_to: int, encryption_key: bytes = None, total_files: int = 1, downloaded_files: int = 0):
    file_url = f"{BASE_URL}/api/download/file/{container_id}/{upload_id}/{signature}/{filename if not encryption_key else 'encrypted'}"
    dest_path = os.path.join(dest, filename)

    if os.path.commonprefix((os.path.realpath(dest_path), dest)) != dest:
        typer.echo(f"Invalid file name: {filename}")
        raise typer.Abort()

    response = requests.get(file_url, stream=True)
    response.raise_for_status()

    contentLength = int(response.headers.get("Content-Length"))
    file_downloaded_size = 0
    iv = None

    with open(dest_path, "wb") as f:
        for chunk in response.iter_content(chunk_size=512 * 1024):
            abort = False
            if encryption_key:
                if not iv:
                    iv = chunk[:16]
                    chunk = chunk[16:]
                chunk = decrypt(encryption_key, iv, chunk)
                if file_downloaded_size == contentLength:
                    chunk = unpad(chunk, AES.block_size)
                file_downloaded_size += len(chunk)
                if file_downloaded_size > file_size:
                    excess = file_downloaded_size - file_size
                    chunk = chunk[:-excess]
                    abort = True
                iv = chunk[-16:]
            else:
                file_downloaded_size += len(chunk)

            f.write(chunk)

            progress = min(file_downloaded_size / file_size, 1)
            status_update_msg = f"Downloading {filename.ljust(pad_to)} [{'█' * int(progress * 20):20s}] {progress:.0%} ({total_files - downloaded_files - 1} left)"
            typer.echo(status_update_msg + "\r", nl=False)

            if abort:
                break

    typer.echo("")
